import numpy as np

from . import register_env
from .half_cheetah import HalfCheetahEnv
import mujoco_py


@register_env("cheetah-vel")
class HalfCheetahVelEnv(HalfCheetahEnv):
    """Half-cheetah environment with target velocity, as described in [1]. The
    code is adapted from
    https://github.com/cbfinn/maml_rl/blob/9c8e2ebd741cb0c7b8bf2d040c4caeeb8e06cc95/rllab/envs/mujoco/half_cheetah_env_rand.py

    The half-cheetah follows the dynamics from MuJoCo [2], and receives at each
    time step a reward composed of a control cost and a penalty equal to the
    difference between its current velocity and the target velocity. The tasks
    are generated by sampling the target velocities from the uniform
    distribution on [0, 2].

    [1] Chelsea Finn, Pieter Abbeel, Sergey Levine, "Model-Agnostic
        Meta-Learning for Fast Adaptation of Deep Networks", 2017
        (https://arxiv.org/abs/1703.03400)
    [2] Emanuel Todorov, Tom Erez, Yuval Tassa, "MuJoCo: A physics engine for
        model-based control", 2012
        (https://homes.cs.washington.edu/~todorov/papers/TodorovIROS12.pdf)
    """

    def __init__(self, task={}, n_tasks=2, randomize_tasks=True):
        self._task = task
        self.tasks = self.sample_tasks(n_tasks)
        self._goal_vel = self.tasks[0].get("velocity", 0.0)
        self._goal = self._goal_vel
        super(HalfCheetahVelEnv, self).__init__()

    def step(self, action):
        xposbefore = self.sim.data.qpos[0]
        self.do_simulation(action, self.frame_skip)
        xposafter = self.sim.data.qpos[0]

        forward_vel = (xposafter - xposbefore) / self.dt
        forward_reward = -1.0 * abs(forward_vel - self._goal_vel)
        ctrl_cost = 0.5 * 1e-1 * np.sum(np.square(action))

        observation = self._get_obs()
        reward = forward_reward - ctrl_cost
        done = False
        infos = dict(
            reward_forward=forward_reward,
            reward_ctrl=-ctrl_cost,
            goal_vel=self._goal_vel,
            forward_vel=forward_vel,
        )
        return (observation, reward, done, infos)

    def reward(self, info, goal):
        reward_ctrl, forward_vel = info["reward_ctrl"], info["forward_vel"]
        forward_reward = -1.0 * abs(forward_vel - goal)
        return forward_reward + reward_ctrl

    def sample_tasks(self, num_tasks):
        np.random.seed(1)
        velocities = np.random.uniform(0.0, 3.0, size=(num_tasks,))
        tasks = [{"velocity": velocity} for velocity in velocities]
        return tasks

    def get_train_goals(self, n_train_tasks):
        return [task["velocity"] for task in self.tasks[:n_train_tasks]]

    def get_all_task_idx(self):
        return range(len(self.tasks))

    def reset_task(self, idx):
        self._task = self.tasks[idx]
        self._goal_vel = self._task["velocity"]
        self._goal = self._goal_vel
        self.reset()

@register_env('cheetah-vel-sparse')
class HalfCheetahVelEnvSparse(HalfCheetahEnv):

    def __init__(self, task={}, n_tasks=2, randomize_tasks=True,goal_radius=0.5):
        self._task = task
        self.tasks = self.sample_tasks(n_tasks)
        self._goal_vel = self.tasks[0].get('velocity', 0.0)
        self._goal = self._goal_vel
        # self.goal_radius= goal_radius
        self.goal_radius = 0.3
        super(HalfCheetahVelEnvSparse, self).__init__()
        self.initialize_camera()

    def step(self, action):
        xposbefore = self.sim.data.qpos[0]
        self.do_simulation(action, self.frame_skip)
        xposafter = self.sim.data.qpos[0]

        forward_vel = (xposafter - xposbefore) / self.dt
        forward_reward = -1.0 * abs(forward_vel - self._goal_vel)
        sparse_reward = self.sparsify_rewards(forward_reward)
        ctrl_cost = 0.5 * 1e-1 * np.sum(np.square(action))

        observation = self._get_obs()
        reward = forward_reward - ctrl_cost
        sparse_reward = sparse_reward - ctrl_cost
        reward = sparse_reward
        done = False
        infos = dict(reward_forward=forward_reward,
            reward_ctrl=-ctrl_cost, task=self._task, velocity=forward_vel,sparse_reward=sparse_reward)
        image = self.get_image()
        return (image, reward, done, infos)

    def initialize_camera(self):
        # set camera parameters for viewing
        sim = self.sim
        viewer = mujoco_py.MjRenderContextOffscreen(sim)
        camera = viewer.cam
        camera.type = 1
        camera.trackbodyid = 0
        camera.elevation = -20
        # camera.azimuth = -90
        sim.add_render_context(viewer)

    def get_image(self, width=100, height=100, camera_name=None):
        image = self.sim.render(
            width=width,
            height=height,
            camera_name=camera_name,
        )
        return np.transpose(image, (2, 0, 1))

    def reward(self, info, goal):
        reward_ctrl, velocity = info["reward_ctrl"], info["velocity"]
        forward_reward = -1.0 * abs(velocity - goal)
        sparse_reward = self.sparsify_rewards(forward_reward)
        return sparse_reward + reward_ctrl, False

    def sample_tasks(self, num_tasks):
        np.random.seed(1)
        velocities = np.random.uniform(0.0, 3.0, size=(num_tasks,))
        tasks = [{'velocity': velocity} for velocity in velocities]
        return tasks

    def sparsify_rewards(self, r):
        ''' zero out rewards when outside the goal radius '''
        #mask = (r >= -self.goal_radius).astype(np.float32)
        #r = r * mask
        if r < - self.goal_radius:
            r = -2
        r = r + 2
        return r

    def get_all_task_idx(self):
        return range(len(self.tasks))

    def get_train_goals(self, n_train_tasks):
        return [task["velocity"] for task in self.tasks[:n_train_tasks]]

    def reset_task(self, idx):
        self._task = self.tasks[idx]
        self._goal_vel = self._task['velocity']
        self._goal = self._goal_vel
        self.reset()

    def reset(self):
        self.reset_model()
        return self.get_image()
